///////////// Copyright © 2016 Dennis Weller. All rights reserved. /////////
//
//   Project     : Xilinx OpenCL Kernels
//   File        : cdot.cl
//   Description :
//      vector scalar product c = x*y
//
//   Created On: 24.05.2016
//   Created By: Dennis Weller
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////



#define REDUCE_FACTOR 2 //reduce number of BRAMS, set to 1 if resource utilization is less than 100%
#define NUMBER_OF_CUS 1 //number of compute units
#define BLOCK_SIZE 1024*1024/4/NUMBER_OF_CUS/REDUCE_FACTOR 
#define BURST_SIZE BLOCK_SIZE/16 //float16

__kernel __attribute__((reqd_work_group_size(1,1,1)))
void kernel_dotc(const int nloops, __global const float16 *x, __global const float16 *y, __global float *out) {
	
	float temp;  
	temp = 0.0f;
	int group_id = get_group_id(0); //for each work_group nloops dot-products are summed up, final reduction on host program

	__local float16 x_local[BURST_SIZE];
	__local float16 y_local[BURST_SIZE];

	ulong blockindex;
    	blockindex = 0;

	for (int loopindex = 0; loopindex < nloops; loopindex++){

	    	__attribute__((xcl_pipeline_loop))
	    	for (blockindex=0; blockindex<BURST_SIZE; blockindex++) {
		   //load data to local memory
	     	   x_local[blockindex] = x[blockindex+(loopindex+group_id*nloops)*BURST_SIZE];
	    	}

            __attribute__((xcl_pipeline_loop))
	    	for (blockindex=0; blockindex<BURST_SIZE; blockindex++) {
		   //load data to local memory
	     	   y_local[blockindex] = y[blockindex+(loopindex+group_id*nloops)*BURST_SIZE];

		   //Carry out computation
		   temp += dot(x_local[blockindex],y_local[blockindex]);
	    	}
	}
	//write result back	
	out[group_id] = temp;
}

