///////////// Copyright © 2016 Dennis Weller. All rights reserved. /////////
//
//   Project     : Xilinx OpenCL Kernels
//   File        : laplace.cl
//   Description :
//      out = A*x, where A is the laplace matrix
//
//   Created On: 27.04.2016
//   Created By: Dennis Weller
////////////////////////////////////////////////////////////////////////////

/**
 * Basic kernel computes 8 rows per kernel using float16
 */

#define RANK 8192
#define NROWS  RANK*RANK //size of unknown vector x, and rank of Laplace matrix
#define X_LOAD_SIZE RANK/16*10 //size of loaded x-vector float16 elements used for computation
#define BLOCKS_PER_ROW RANK/16  //float16

__kernel __attribute__((reqd_work_group_size(1,1,1)))
void kernel_laplace_scalar(__global const float16 *x,  __global float16 *out) {

	int gid = get_global_id(0);
	int group_id = get_group_id(0);
	int offset = group_id;
	local float16 x_local[X_LOAD_SIZE];
	local float16 out_local[BLOCKS_PER_ROW*8]; //compute 8 rows per kernel
				
	int row_index;
	int blockgroup;
	int block_index;
	int x_index;	
	row_index = 0;	
	blockgroup = 0;
	block_index = 0;
	x_index = 0;

	float temp_arr[16];

	local float16 laplace_vec1[16];
	laplace_vec1[0] = (float16)	( 4.0f ,-1.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ) ;
	laplace_vec1[1] = (float16)	(-1.0f , 4.0f ,-1.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ) ;
	laplace_vec1[2] = (float16)	( 0.0f ,-1.0f , 4.0f ,-1.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ) ;
	laplace_vec1[3] = (float16)	( 0.0f , 0.0f ,-1.0f , 4.0f ,-1.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ) ;
	laplace_vec1[4] = (float16)	( 0.0f , 0.0f , 0.0f ,-1.0f , 4.0f ,-1.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ) ;
	laplace_vec1[5] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f ,-1.0f , 4.0f ,-1.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ) ;
	laplace_vec1[6] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,-1.0f , 4.0f ,-1.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ) ;
	laplace_vec1[7] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,-1.0f , 4.0f , -1.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ) ;
	laplace_vec1[8] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,-1.0f ,  4.0f ,-1.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ) ;
	laplace_vec1[9] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , -1.0f , 4.0f ,-1.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ) ;
	laplace_vec1[10] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f ,-1.0f , 4.0f ,-1.0f , 0.0f , 0.0f , 0.0f , 0.0f ) ;
	laplace_vec1[11] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f ,-1.0f , 4.0f ,-1.0f , 0.0f , 0.0f , 0.0f ) ;
	laplace_vec1[12] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f ,-1.0f , 4.0f ,-1.0f , 0.0f , 0.0f ) ;
	laplace_vec1[13] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f ,-1.0f , 4.0f ,-1.0f , 0.0f ) ;
	laplace_vec1[14] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,-1.0f , 4.0f ,-1.0f ) ;
	laplace_vec1[15] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,-1.0f , 4.0f ) ; 
					
					
	local float16 laplace_vec2[16];
 	laplace_vec2[0] = (float16) 	( -1.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f );
	laplace_vec2[1] = (float16)	( 0.0f , -1.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f );
	laplace_vec2[2] = (float16)	( 0.0f , 0.0f , -1.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f );
	laplace_vec2[3] = (float16)	( 0.0f , 0.0f , 0.0f , -1.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f );
	laplace_vec2[4] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , -1.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f );
	laplace_vec2[5] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , -1.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f );
	laplace_vec2[6] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , -1.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f );
	laplace_vec2[7] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , -1.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f );
	laplace_vec2[8] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  -1.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f );
	laplace_vec2[9] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , -1.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f );
	laplace_vec2[10] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , -1.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f );
	laplace_vec2[11] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , -1.0f , 0.0f , 0.0f , 0.0f , 0.0f );
	laplace_vec2[12] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , -1.0f , 0.0f , 0.0f , 0.0f );
	laplace_vec2[13] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , -1.0f , 0.0f , 0.0f );
	laplace_vec2[14] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , -1.0f , 0.0f );
	laplace_vec2[15] = (float16)	( 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f ,  0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , 0.0f , -1.0f );
	
	int x_past_row;
	int x_current_row;
	int x_next_row;

	
	//Burst transfer: load 10 rows each of size RANK 
	__attribute__((xcl_pipeline_loop))
	loopa:
    	for (x_index=0; x_index<X_LOAD_SIZE; x_index++) 
     	   x_local[x_index] = x[x_index+offset*BLOCKS_PER_ROW*8];
	
	//Processing of 8 rows
	loopb:
	for( row_index=0 ; row_index <  8; row_index++){ 
		
		x_past_row = row_index * BLOCKS_PER_ROW; 
		x_current_row = x_past_row + BLOCKS_PER_ROW;  
		x_next_row = x_current_row + BLOCKS_PER_ROW; 
			    	

	    //1st blockgroup

		//1st element at the left border
		temp_arr[0] = dot( laplace_vec1[0], x_local[x_current_row]) + dot(laplace_vec2[0], x_local[x_past_row]) + dot(laplace_vec2[0] , x_local[x_next_row]);
		
		//middle elements
		__attribute__((xcl_pipeline_loop))		
		loopb_a:
		for ( block_index = 1; block_index < 15; block_index++)
			temp_arr[block_index] = dot(laplace_vec1[block_index], x_local[x_current_row]) + dot(laplace_vec2[block_index], x_local[x_past_row]) + dot(laplace_vec2[block_index] , x_local[x_next_row]);
		
		//16th element
		temp_arr[15] = dot(laplace_vec1[15], x_local[x_current_row]) + dot(laplace_vec2[0], x_local[x_current_row+1]) + dot(laplace_vec2[15], x_local[x_past_row]) + dot(laplace_vec2[15] , x_local[x_next_row]);

		//Package all elements in a float16 container
		out_local[row_index*BLOCKS_PER_ROW] = (float16) (temp_arr[0], temp_arr[1],temp_arr[2],temp_arr[3], temp_arr[4], temp_arr[5],temp_arr[6],temp_arr[7],temp_arr[8], temp_arr[9],temp_arr[10],temp_arr[11], temp_arr[12], temp_arr[13],temp_arr[14],temp_arr[15]);
	
		
	    //2nd to last-1 blockgroup
		__attribute__((xcl_pipeline_loop))
		loopb_b:
		for (blockgroup = 1; blockgroup < BLOCKS_PER_ROW-1; blockgroup++){
			
			//1st element			
			temp_arr[0] = dot(laplace_vec1[0], x_local[x_current_row+blockgroup]) + dot(laplace_vec2[15], x_local[x_current_row+blockgroup-1]) + dot(laplace_vec2[0], x_local[x_past_row+blockgroup]) + dot(laplace_vec2[0] , x_local[x_next_row + blockgroup]);
			
			//middle elements		
			loopb_b_a:			
			for ( block_index = 1; block_index < 15 ; block_index++){
				temp_arr[block_index] = dot(laplace_vec1[block_index], x_local[x_current_row+blockgroup]) +  dot(laplace_vec2[block_index], x_local[x_past_row + blockgroup]) + dot(laplace_vec2[block_index] , x_local[x_next_row + blockgroup]);
			}

			//16th element
			temp_arr[15] = dot(laplace_vec1[15], x_local[x_current_row+blockgroup]) + dot(laplace_vec2[0], x_local[x_current_row+blockgroup+1]) + dot(laplace_vec2[15], x_local[x_past_row + blockgroup]) + dot(laplace_vec2[15] , x_local[x_next_row + blockgroup]);

			//Package all elements in a float16 container
			out_local[row_index*BLOCKS_PER_ROW+blockgroup] = (float16) (temp_arr[0], temp_arr[1],temp_arr[2],temp_arr[3], temp_arr[4], temp_arr[5],temp_arr[6],temp_arr[7],temp_arr[8], temp_arr[9],temp_arr[10],temp_arr[11], temp_arr[12], temp_arr[13],temp_arr[14],temp_arr[15]);
		}

	    //last blockgroup
		
		//1st element
		temp_arr[0] = dot(laplace_vec1[0], x_local[x_current_row+BLOCKS_PER_ROW-1]) + dot(laplace_vec2[15], x_local[x_current_row+BLOCKS_PER_ROW-2]) + dot(laplace_vec2[0], x_local[x_past_row+BLOCKS_PER_ROW-1]) + dot(laplace_vec2[0] , x_local[x_next_row + BLOCKS_PER_ROW-1]);
		
		//middle elements
		__attribute__((xcl_pipeline_loop))		
		loopb_c:
		for ( block_index = 1; block_index < 15 ; block_index++){
			temp_arr[block_index] = dot(laplace_vec1[block_index], x_local[x_current_row+BLOCKS_PER_ROW-1]) +  dot(laplace_vec2[block_index], x_local[x_past_row+BLOCKS_PER_ROW-1]) + dot(laplace_vec2[block_index] , x_local[x_next_row + BLOCKS_PER_ROW-1]);
		}

		//16th element at the right border
		temp_arr[15] = dot(laplace_vec1[15], x_local[x_current_row+BLOCKS_PER_ROW-1]) +  dot(laplace_vec2[15], x_local[x_past_row+BLOCKS_PER_ROW-1]) + dot(laplace_vec2[15] , x_local[x_next_row + BLOCKS_PER_ROW-1]);
		
		//Package all elements in a float16 container
		out_local[row_index*BLOCKS_PER_ROW+BLOCKS_PER_ROW-1] = (float16) (temp_arr[0], temp_arr[1],temp_arr[2],temp_arr[3], temp_arr[4], temp_arr[5],temp_arr[6],temp_arr[7],temp_arr[8], temp_arr[9],temp_arr[10],temp_arr[11], temp_arr[12], temp_arr[13],temp_arr[14],temp_arr[15]);

		
	}

	//write back the result of 8 rows
	__attribute__((xcl_pipeline_loop))
	loopc:
    	for (blockgroup=0; blockgroup < BLOCKS_PER_ROW*8; blockgroup++) {
     	   out[offset*8*BLOCKS_PER_ROW + blockgroup] = out_local[blockgroup];
    	}


}	
		

