///////////// Copyright © 2016 Dennis Weller. All rights reserved. /////////
//
//   Project     : Xilinx OpenCL Kernels
//   File        : vectorAdd.cl
//   Description :
//      vector addition z = alpha*x+y
//
//   Created On: 24.05.2016
//   Created By: Dennis Weller
////////////////////////////////////////////////////////////////////////////

#define REDUCE_FACTOR 8 //reduce number of BRAMS, can set to 1 if utilization is less than 100%
#define BLOCK 1024*1024/4/REDUCE_FACTOR 
#define BURST_SIZE BLOCK/16 //float16

__kernel __attribute__((reqd_work_group_size(1,1,1)))
void kernel_scaleadd(const int nloops, const float alpha,__global const float16 *x,__global const float16 *y, __global float16 *out) 
{	
	__local float16 x_local[BURST_SIZE];
	__local float16 y_local[BURST_SIZE];
	__local float16 out_local[BURST_SIZE];

	for( int loopindex = 0; loopindex < nloops; loopindex++){
	    	
	    	ulong blockindex;
	    	blockindex = 0;
	    	__attribute__((xcl_pipeline_loop))
	    	for (blockindex=0; blockindex<BURST_SIZE; blockindex++) {
		   //read data and save to local memory
	     	   x_local[blockindex] = x[blockindex+loopindex*BURST_SIZE];
	    	}

		__attribute__((xcl_pipeline_loop))
	    	for (blockindex=0; blockindex<BURST_SIZE; blockindex++) {
		   //read data and save to local memory
	     	   y_local[blockindex] = y[blockindex+loopindex*BURST_SIZE];
	    	}
	

		__attribute__((xcl_pipeline_loop))
	    	for (blockindex=0; blockindex<BURST_SIZE; blockindex++) {
		   //compute scaleadd and save to local memory
		   out_local[blockindex] = (x_local[blockindex] * alpha) + y_local[blockindex];

		   //write back result to global memory
	     	   out[blockindex+loopindex*BURST_SIZE] = out_local[blockindex];
	    	}
	}
}

